@@ -7,9 +7,11 @@ use Carp;
 use JSON::XS qw//;
 
 use Furl;
+use File::Spec;
 use URI;
+use Try::Tiny;
 
-our $VERSION = "0.6";
+our $VERSION = "0.7";
 
 our %OPT_TABLE = ();
 
@@ -25,6 +27,8 @@ sub new {
         doas => $opts{doas},
         useragent => $opts{useragent} || 'Furl Net::Hadoop::WebHDFS (perl)',
         timeout => $opts{timeout} || 10,
+        suppress_errors => $opts{suppress_errors} || 0,
+        last_error => undef,
         under_failover => 0,
     };
     $self->{furl} = Furl::HTTP->new(agent => $self->{useragent}, timeout => $self->{timeout}, max_redirects => 0);
@@ -305,13 +309,25 @@ sub operate_requests {
     return $self->request($uri->host, $uri->port, $method, $uri->path_query, undef, {}, $payload, $headers);
 }
 
+sub request {
+    my $self = shift;
+    return $self->_request(@_) unless $self->{suppress_errors};
+
+    try {
+        $self->_request(@_);
+    } catch {
+        $self->{last_error} = $_;
+        0;
+    };
+}
+
 # IllegalArgumentException      400 Bad Request
 # UnsupportedOperationException 400 Bad Request
 # SecurityException             401 Unauthorized
 # IOException                   403 Forbidden
 # FileNotFoundException         404 Not Found
 # RumtimeException              500 Internal Server Error
-sub request {
+sub _request {
     my ($self, $host, $port, $method, $path, $op, $params, $payload, $header) = @_;
 
     my $request_path = $op ? $self->build_path($path, $op, %$params) : $path;
@@ -366,6 +382,93 @@ sub request {
     croak "RequestFailedError, code:$code, message:$errmsg";
 }
 
+sub exists {
+    my $self = shift;
+    my $path = shift || croak "No HDFS path was specified";
+    my $stat;
+    eval {
+        $stat = $self->stat( $path );
+        1;
+    } or do {
+        my $eval_error = $@ || 'Zombie error';
+        return if $eval_error =~ m<
+            \QFileNotFoundError: {"RemoteException":{"message":"File does not exist:\E
+        >xms;
+        # just re-throw
+        croak $eval_error;
+    };
+    return $stat;
+}
+
+sub find {
+    my $self      = shift;
+    my $file_path = shift || croak "No file path specified";
+    my $cb        = shift;
+    my $opt       = @_ && ref $_[-1] eq 'HASH' ? pop @_ : {};
+
+    if ( ref $cb ne 'CODE' ) {
+        die "Call back needs to be a CODE ref";
+    }
+
+    my $suppress = $self->{suppress_errors};
+    # can be used to quickly skip the java junk like, file names starting with
+    # underscores, etc.
+    my $re_ignore     = $opt->{re_ignore} ? qr/$opt->{re_ignore}/ : undef;
+
+    #
+    # No such thing like symlinks (yet) in HDFS, in case you're wondering:
+    # https://issues.apache.org/jira/browse/HADOOP-10019
+    # although check that link yourself
+    #
+    my $looper;
+    $looper = sub {
+        my $thing = shift;
+        if ( ! $self->exists( $thing ) ) {
+            # should happen at the start, so this will short-circuit the recursion
+            warn "The HDFS directory specified ($thing) does not exist! Please guard your HDFS paths with exists()";
+            return;
+        }
+        my $list = $self->list( $thing );
+        foreach my $e ( @{ $list } ) {
+            my $path = $e->{pathSuffix};
+            my $type = $e->{type};
+
+            next if $re_ignore && $path && $path =~ $re_ignore;
+
+            if ( $type eq 'DIRECTORY' ) {
+                $cb->( $thing, $e );
+                eval {
+                    $looper->( File::Spec->catdir( $thing, $path ) );
+                    1;
+                } or do {
+                    my $eval_error = $@ || 'Zombie error';
+                    if ( $suppress ) {
+                        warn "[ERROR DOWNGRADED] Failed to check $thing/$path: $eval_error";
+                        next;
+                    }
+                    croak $eval_error;
+                }
+            }
+            elsif ( $type eq 'FILE' ) {
+                $cb->( $thing, $e );
+            }
+            else {
+                my $msg = "I don't know what to do with type=$type!";
+                if ( $suppress ) {
+                    warn "[ERROR DOWNGRADED] $msg";
+                    next;
+                }
+                croak $msg;
+            }
+        }
+        return;
+    };
+
+    $looper->( $file_path );
+
+    return;
+}
+
 1;
 
 __END__
@@ -399,7 +502,7 @@ Net::Hadoop::WebHDFS class method and instance methods.
 =head3 C<< Net::Hadoop::WebHDFS->new( %args ) :Net::Hadoop::WebHDFS >>
 
 Creates and returns a new client instance with I<%args>.
-If you are using HttpFs, set I<httpfs_mode => 1> and I<port => 14000>.
+If you are using HttpFs, set I<< httpfs_mode => 1 >> and I<< port => 14000 >>.
 
 
 I<%args> might be:
@@ -482,7 +585,7 @@ Rename file or directory as I<$dest>.
 
 =head3 C<< $client->delete($path, [recursive => 0/1]) :Bool >>
 
-Delete file I<$path> from HDFS. With optional I<recursive => 1>, files and directories are removed recursively (default false).
+Delete file I<$path> from HDFS. With optional I<< recursive => 1 >>, files and directories are removed recursively (default false).
 
 =head3 C<< $client->stat($path) :HashRef >>
 
@@ -524,6 +627,37 @@ Set mtime/atime of I<$path>. Alias: B<settimes>.
 
 Create a zero length file.
 
+=head2 EXTENSIONS
+
+=head3 C<< $client->exists($path) :Bool >>
+
+Returns the C<< stat() >> hash if successful, and false otherwise. Dies on
+interface errors.
+
+=head3 C<< $client->find($path, $callback, $options_hash) >>
+
+Loops recursively over the specified path:
+
+    $client->find(
+        '/user',
+        sub {
+            my($cwd, $path) = @_;
+            my $date = localtime $path->{modificationTime};
+            my $type = $path->{type} eq q{DIRECTORY} ? "[dir ]" : "[file]";
+            my $size = sprintf "% 10s",
+                                $path->{blockSize}
+                                    ? sprintf "%.2f MB", $path->{blockSize} / 1024**2
+                                    : 0;
+            print "$type $size $path->{permission} $path->{owner}:$path->{group} $cwd/$path->{pathSuffix}\n";
+        },
+        { # optional
+            re_ignore => qr{
+                            \A      # Filter some filenames out even before reaching the callback
+                                [_] # logs and meta data, java junk, _SUCCESS files, etc.
+                        }xms,
+        }
+    );
+
 =head1 AUTHOR
 
 TAGOMORI Satoshi E<lt>tagomoris {at} gmail.comE<gt>
